/*RSBAC global network
 *
 *(c) Stanislav Ievlev <inger@linux.ru.net>
 */

/*r-jobd.c - jobs server*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>

#include <security/pam_appl.h>
#include <security/pam_misc.h>

#include "r-job.h"
#include "r-job-d.h"
#include "auth.h"
#include "rsa.h"

/* conversation function & corresponding structure */
static struct pam_conv conv={0,0};

/*keywords*/
char crypto_keyword[MAX_LINE],session_keyword[MAX_LINE];
struct sockaddr_in serv_addr, clnt_addr; /*addres structures*/

/*for commands*/
int add_only=0;
int get_only=0;
int write_mode=0; /*send data mode*/
int read_mode=0; /*recieve data mode*/
int server_pause=0;    /*pause mode*/
int warranty=0;  /*warranty for job receiving*/

char request[MAX_LINE]; /*temporally request field*/

/*for future use*/
char *hostname=NULL; /*slave host*/
char *clientlist_filename=NULL; /*station list file*/
char *job_filename=NULL; /*pattern for job-file name*/
char *log_filename=NULL; /*log file name*/
char *errlog_filename=NULL; /*err log file name*/

char *config_filename=NULL;/*configuration file*/

int  network=0;

/*forward declarations*/
void stop_work(int sig);
extern void fd_log_start(char *station,int level);
extern void fd_log(char *command_str);
extern void fd_log_arg(char *command_str);
extern void fd_log_end();
extern void fd_err_log(char *err_str,char *station);

/*PAM conversation function*/
static int stdin_conv(int num_msg, const struct pam_message **msgm,
		      struct pam_response **response, void *appdata_ptr) {
    struct pam_response *reply;
    int count;

    if (num_msg <= 0)
	return PAM_CONV_ERR;

    reply = (struct pam_response *) calloc(num_msg,
					   sizeof(struct pam_response));
    if (reply == NULL) {
	return PAM_CONV_ERR;
    }

    for (count=0; count < num_msg; ++count) {
	reply[count].resp_retcode = 0;
	reply[count].resp = strdup(appdata_ptr);
    }

    *response = reply;
    reply = NULL;

    return PAM_SUCCESS;
}


/*read and analise command from client*/
int read_command(int ns)
{
    char command_str[MAX_LINE];
    int res=0;
    int level=0;
    
    
    /*set default values*/
    get_only=0;
    add_only=0;
    read_mode=0;
    write_mode=0;
    warranty=0;
    
    /*get station's level*/
    level=get_station_level(inet_ntoa(clnt_addr.sin_addr));
    
    if (recv(ns, command_str, MAX_LINE, 0)==-1){
	syslog(LOG_INFO, "rsbac-net: Cannot recieve message.");
        }

#ifdef TEST
    printf(">>command:%s\n",command_str);
    printf(">>station_level:%u\n",level);
#endif

    /*logging facility*/
    fd_log_start(inet_ntoa(clnt_addr.sin_addr),level);
    fd_log(command_str);
    
    /*Parse commands*/
    
    /*High level commands*/
if (level==MAX_LEVEL){
    /*May be stop server?*/
    if (!(strncmp(command_str,STOP_SERVER_COMMAND,MAX_LINE))){
	fd_log_end();
	stop_work(0);
	res=1;
	}
	
    /*May be pause server?*/
    if (!(strncmp(command_str,PAUSE_SERVER_COMMAND,MAX_LINE))){
	server_pause=1;
	get_only=1;
	add_only=1;
	res=1;
    }

    /*May be resume server?*/
    if (!(strncmp(command_str,RESUME_SERVER_COMMAND,MAX_LINE))){
	server_pause=0;
	get_only=1;
	add_only=1;
	res=1;
    }

    /*May be clean all jobs?*/
    if (!(strncmp(command_str,CLEAN_COMMAND,MAX_LINE))){
	clean_all_job();
	get_only=1;
	add_only=1;
	res=1;
    }

    /*May be add new station to list?*/
    if (!(strncmp(command_str,ADD_STATION_COMMAND,MAX_LINE))){
	add_only=1;
	read_mode=R_ADD_STATION;
	res=1;
    }

    /*May be delete some station from list?*/
    if (!(strncmp(command_str,DEL_STATION_COMMAND,MAX_LINE))){
	add_only=1;
	read_mode=R_DEL_STATION;
	res=1;
    }
    
    
    /*May be clean all jobs to selected station?*/
    if (!(strncmp(command_str,CLEAN_ONE_STATION_COMMAND,MAX_LINE))){
	add_only=1;
	read_mode=R_CLEAN_ONE;
	res=1;
    }
    
}/*end of high level commands*/

if (level>=MED2_LEVEL){/*Medium level 2 commands*/
    /*May be add job to selected station*/
    if (!(strncmp(command_str,ADD_ONLY_ONE_STATION_COMMAND,MAX_LINE))){
	add_only=1;
	read_mode=R_ADD_ONLY;
	res=1;
    }
    /*May be peek only?*/
    if (!(strncmp(command_str,PEEK_ONE_STATION_COMMAND,MAX_LINE))){
	add_only=0;
	get_only=0;
	read_mode=R_PEEK_ONE;	
	write_mode=W_PEEK_ONE;	
	res=1;
	};
	
    /*May be get jobs from selected station*/
    if (!(strncmp(command_str,GET_ONLY_ONE_STATION_COMMAND,MAX_LINE))){
	add_only=0;
	get_only=0;
	read_mode=R_PEEK_ONE; /*Work similar peek_one_station*/
	write_mode=W_ADD_ONE;
	res=1;
    }


}/*end of medium 2 level commands*/

if (level>=MED1_LEVEL){ /*Medium level 1 commands*/
    /*May be add only?*/
    if (!(strncmp(command_str,ADD_ONLY_COMMAND,MAX_LINE))){
	add_only=1;
	res=1;
	}
	
    /*May be add job to myself?*/
    if (!(strncmp(command_str,ADD_ONLY_MINE_COMMAND,MAX_LINE))){
	add_only=1;
	read_mode=R_ADD_MINE;
	res=1;
    }

    /*May be add and get?*/
    if (!(strncmp(command_str,GET_AND_ADD_COMMAND,MAX_LINE))){
	get_only=0;
	add_only=0;
	res=1;
    }
    /*May be Send with warranty?*/
    if (!(strncmp(command_str,GET_AND_ADD_WARRANTY_COMMAND,MAX_LINE))){
	get_only=0;
	add_only=0;
	warranty=1;
	res=1;
    }
    
} /*end of medium level commands*/

if (level>=MIN_LEVEL){ /*Everybody granted actions*/

    /*May be get only?*/
    if (!(strncmp(command_str,GET_ONLY_COMMAND,MAX_LINE))){
	get_only=1;
	res=1;
	};
	
    /*May be send with warranty?*/
    if (!(strncmp(command_str,GET_ONLY_WARRANTY_COMMAND,MAX_LINE))){
	get_only=1;
	warranty=1;
	res=1;
	};
    
    /*May be see list of allowed stations?*/
    if (!(strncmp(command_str,LIST_STATION_COMMAND,MAX_LINE))){
	get_only=1;
	write_mode=W_LIST_STATION;
	res=1;
    }
    
} /*end of everybody granted actions*/    


    return res;
}

/*read message from client*/
void server_read_message(int ns,char  *session_keyword,char *crypto_keyword)
{
    char command_str[MAX_LINE];
    
    if (recv(ns, command_str, MAX_LINE, 0)==-1){
	syslog(LOG_INFO, "rsbac-net: Cannot recieve message.");
        }/*receive message*/
    my_decrypt(session_keyword,command_str);/*and decrypt*/
	
#ifdef TEST
    printf(">>recieve:%s\n",command_str);
#endif
    
    /*actions*/
    switch (read_mode) {
    case 0: my_encrypt(crypto_keyword,command_str);/*re-crypt using local key*/
	    append_job(command_str);
	    break;
    case R_ADD_STATION: fd_log_arg(command_str); append_station(command_str); 
	    break;
    case R_DEL_STATION: fd_log_arg(command_str); remove_station(command_str);
	    break;
    case R_ADD_ONLY: append_job_to_station(ns,command_str,session_keyword,crypto_keyword);
	    break;
    case R_CLEAN_ONE: clean_all_job_to_station(command_str);
	    break;
    case R_PEEK_ONE:strncpy(request,command_str,MAX_LINE);
	    break;
    case R_ADD_MINE: my_encrypt(crypto_keyword,command_str);/*re-crypt using local key*/
		     append_job_mine(command_str);
		     break;


    default:break;
    }
}

/*write message to client*/
void server_write_message(int ns,char *session_keyword,char *crypto_keyword)
{
    switch (write_mode){
        case 0: send_jobs_station(ns,session_keyword,crypto_keyword,warranty,inet_ntoa(clnt_addr.sin_addr));
		break;
        case W_LIST_STATION: send_stations(ns,session_keyword,crypto_keyword);
		break;
        case W_PEEK_ONE: peek_jobs(ns,session_keyword,crypto_keyword,request);
		break;
        case W_ADD_ONE: send_jobs_station(ns,session_keyword,crypto_keyword,warranty,request);
		break;

	default:break;
    }/*switch*/    
}

/*get password from user*/
void get_keyword()
{
    char *tmp_keyword;
    
    tmp_keyword=getpass("Enter crypto key:");
    strncpy((char *)crypto_keyword,tmp_keyword,MAX_LINE);

#ifdef TEST
    printf(">>crypto:%s\n",crypto_keyword);
#endif
}

/*check password*/
int server_check_keyword(int ns)
{
    int retval;
    char client_keyword[MAX_LINE];
    char log_str[MAX_LINE];
    pam_handle_t *pamh=NULL;
    
    /*in station list?*/
    if (!(in_station_list(inet_ntoa(clnt_addr.sin_addr)))) {
    fd_err_log("FAILED: IP is not in list.",inet_ntoa(clnt_addr.sin_addr));
    return 0;
    }
	
    /*RSA*/
    server_accept_station(ns,inet_ntoa(clnt_addr.sin_addr));
    if (server_test(ns,inet_ntoa(clnt_addr.sin_addr))==1)
    {
#ifdef TEST
    fprintf(stderr,">>RSA:ok\n");
#endif
    
    }else{
#ifdef TEST
    fprintf(stderr,">>RSA:failed\n");
#endif
    fd_err_log("FAILED: RSA",inet_ntoa(clnt_addr.sin_addr));
    return 0;
    }
    
    /* Ok it is our station. We can continue */
    /*Diffie-Hellman key exchange*/
    DH_check(ns,(char *)session_keyword,1);
#ifdef TEST
    fprintf(stderr,">>session keyword:%s\n",session_keyword);
#endif

    /*Password checking throught PAM*/
    if (recv(ns, (char *)client_keyword, MAX_LINE, 0)==-1){/*recieve password*/
	syslog(LOG_INFO, "rsbac-net: Cannot receive password.");
	fd_err_log("FAILED: Cannot receive password.",inet_ntoa(clnt_addr.sin_addr));
        return 0;
        }
    my_decrypt(session_keyword,(char *)client_keyword);/*then decrypt*/

    conv.conv = stdin_conv;
    conv.appdata_ptr = strdup(client_keyword);
    retval = pam_start("r-job", "root", &conv, &pamh);
    
    while (retval == PAM_SUCCESS) {/* use loop to avoid goto... */
	retval = pam_authenticate(pamh, 0);
	
	if (retval != PAM_SUCCESS)
	    break;
	    
	/* all done */
	retval = pam_end(pamh, PAM_SUCCESS);
	if (retval != PAM_SUCCESS)
	    break;
	    
	/* quit gracefully */
#ifdef TEST
	fprintf(stderr,">>password ok\n");
#endif
	break;
    }

    if (retval != PAM_SUCCESS){
#ifdef TEST
	fprintf(stderr, ">>password error: %s\n", pam_strerror(pamh, retval));
#endif
	snprintf(log_str,MAX_LINE,"FAILED: %s",pam_strerror(pamh, retval));
	fd_err_log(log_str,inet_ntoa(clnt_addr.sin_addr));
	
	if (pamh != NULL) {
	    (void) pam_end(pamh,PAM_SUCCESS);
	    pamh = NULL;
	}
    return 0;
    }
	    
    return 1;
}

/*stop all works*/
void stop_work(int sig)
{
    /*send message to log*/
    syslog(LOG_INFO,"rsbac-net: RSBAC network server stopped");
    exit (0);
}

/*main program*/
int main(int argc,char *argv[])
{
    int ch,i; /*options*/
    int ns; /*client connection socket*/
    int s; /*server connection socket*/




    /*check args*/
    while ((ch = getopt(argc, argv, "c:")) != EOF)
	switch((char)ch)
	    {
		case 'c':		/* Station list*/
		    config_filename = optarg;
		    break;
	    }
	
    /*set default values if we need*/
    
    if (config_filename==NULL) /*station list*/
    {
    config_filename=malloc(MAX_LINE);
    strncpy(config_filename,CONFIG_FILE_PATH,MAX_LINE);
    }
    
    /*parse config file first*/
    parse_config(config_filename);
    
#ifdef TEST
    fprintf(stderr,">>station-list:%s\n",clientlist_filename);
    fprintf(stderr,">>job-name:%s\n",job_filename);
    fprintf(stderr,">>activity-log:%s\n",log_filename);
    fprintf(stderr,">>error-log:%s\n",errlog_filename);
#endif

/*get password*/
    get_keyword();
/*open log*/
    openlog("rsbac-netd",LOG_PID,LOG_AUTH);
/*create socket*/
    if((s=socket(AF_INET, SOCK_STREAM, 0))==-1){
	perror("socket()");
	exit(1);
    }
/*fill address structure*/
    bzero(&serv_addr, sizeof(serv_addr));
    serv_addr.sin_family=AF_INET;
    serv_addr.sin_addr.s_addr=INADDR_ANY;
    serv_addr.sin_port=htons(PORTNUM);
    
    if( bind(s, (struct sockaddr_in *)&serv_addr, sizeof(serv_addr))==-1){
	perror("bind()");
	exit(1);
    }

    if(listen(s,5)==-1){
	perror("listen()");
        exit(1);
    }
    
/*go to daemon mode*/
    signal(SIGINT, stop_work); 
#ifndef TEST
    /*ignore all signals.*/
    for (i = 1; i < NSIG; i++)	signal(i, SIG_IGN); 


    signal(SIGKILL, stop_work);
    signal(SIGTERM, stop_work);
    
    /*make init father*/
    if ( fork() == 0 )
    {
	syslog(LOG_INFO,"rsbac-net: RSBAC network  server started");
    }else{
	exit(0); /*father exits*/
    }
    
    /*setup new session*/
    setsid();
#endif

/*working circle*/
    while (1){
	int addrlen=sizeof(clnt_addr);
	
	bzero(&clnt_addr,sizeof(clnt_addr));

	if((ns=accept(s, (struct sockaddr_in *)&clnt_addr,&addrlen))==-1){
	    /*perror("accept()");*/
	    syslog(LOG_INFO, "rsbac-netd: Cannot accept connection.");    
	    /*exit(1);*/
	}
/*process for client*/

	/*check password*/
	if (server_check_keyword(ns)){
	    /*recieving command*/
	    if (read_command(ns)==1){
	        /*work if no pause*/
		if (server_pause==0){
    		    /*recieving data*/
		    if(get_only==0)  server_read_message(ns,session_keyword,crypto_keyword);
		    /*sending data*/
		    if(add_only==0)  server_write_message(ns,session_keyword,crypto_keyword);
		
    		}
		fd_log_end();
	    }else{/*wrong command*/
#ifdef TEST
	    printf(">>invalid command\n");
#endif
	    }
	}/*if*/
	
	close(ns);
    }/*while*/

/*close server*/
close(s);  
 
return (0);
}	